/*
** File: admxrc3d.c  
** Project: ADMXRC3 API module library
** Purpose: OS-independent code for ADMXRC3 API module library, sensor functions
**
** (C) Copyright Alpha Data 2010-2011
*/

#include <admxrc3.h>
#include <admxrc3_platform.h>
#include <portable_file.h>
#include <portable_string.h>
#if PORTABLE_HAS_WCHAR_SUPPORT
# include <portable_wstring.h>
#endif
#include "ioctl_admxrc3.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static const double g_64k = 65536.0;

#define TO_DOUBLE(i, f) ((double)(i) + (double)(f) / g_64k);

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetSensorInfoA(
  ADMXRC3_HANDLE hCard,
  unsigned int sensorIndex,
  ADMXRC3_SENSOR_INFOA* pSensorInfo)
{
  IOCTLS_ADMXRC3_GETSENSORINFO ioctl;
  ADMXRC3_STATUS status;

  if (NULL == pSensorInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.sensorIndex = sensorIndex;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_GETSENSORINFO, &ioctl, sizeof(ioctl.in), sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    memset(pSensorInfo, 0, sizeof(*pSensorInfo));
    portableStringCopy(pSensorInfo->Description, ARRAY_LENGTH(pSensorInfo->Description), ioctl.out.description);
    pSensorInfo->Capabilities = ioctl.out.capabilities;
    pSensorInfo->Error = TO_DOUBLE(ioctl.out.error.intPart, ioctl.out.error.fracPart);
    pSensorInfo->DataType = ioctl.out.datatype;
    pSensorInfo->Unit = ioctl.out.unit;
    pSensorInfo->Exponent = ioctl.out.exponent;
  }
  return status;
}

#if PORTABLE_HAS_WCHAR_SUPPORT
ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_GetSensorInfoW(
  ADMXRC3_HANDLE hCard,
  unsigned int sensorIndex,
  ADMXRC3_SENSOR_INFOW* pSensorInfo)
{
  ADMXRC3_SENSOR_INFOA sensorInfoA;
  ADMXRC3_STATUS status;

  if (NULL == pSensorInfo) {
    return ADMXRC3_NULL_POINTER;
  }

  status = ADMXRC3_GetSensorInfoA(hCard, sensorIndex, &sensorInfoA);
  if (ADMXRC3_SUCCESS == status) {
    portableUTF8ToWideChar(pSensorInfo->Description, ARRAY_LENGTH(pSensorInfo->Description), sensorInfoA.Description);
    pSensorInfo->Capabilities = sensorInfoA.Capabilities;
    pSensorInfo->Error = sensorInfoA.Error;
    pSensorInfo->DataType = sensorInfoA.DataType;
    pSensorInfo->Unit = sensorInfoA.Unit;
    pSensorInfo->Exponent = sensorInfoA.Exponent;
  }
  return status;
}
#endif

ADMXRC3_EXPORT ADMXRC3_STATUS
ADMXRC3_CALLING_CONVENTION
ADMXRC3_ReadSensor(
  ADMXRC3_HANDLE hCard,
  unsigned int sensorIndex,
  ADMXRC3_SENSOR_VALUE* pSensorValue)
{
  IOCTLS_ADMXRC3_READSENSORNOTYPE ioctl;
  ADMXRC3_STATUS status;

  if (NULL == pSensorValue) {
    return ADMXRC3_NULL_POINTER;
  }

  ioctl.in.sensorIndex = sensorIndex;
  status = admxrc3Ioctl(hCard, ADMXRC3_IOCTLCODE_READSENSORNOTYPE, &ioctl, sizeof(ioctl.in), sizeof(ioctl.out));
  if (ADMXRC3_SUCCESS == status) {
    switch (ioctl.out.datatype) {
    case ADMXRC3_DATA_BOOL:
      pSensorValue->Bool = ioctl.out.value.booleanValue;
      break;

    case ADMXRC3_DATA_DOUBLE:
      pSensorValue->Double = TO_DOUBLE(ioctl.out.value.doubleValue.intPart, ioctl.out.value.doubleValue.fracPart);
      break;

    case ADMXRC3_DATA_INT32:
      pSensorValue->Int32 = ioctl.out.value.int32Value;
      break;

    case ADMXRC3_DATA_UINT32:
      pSensorValue->UInt32 = ioctl.out.value.uint32Value;
      break;

    default:
      return ADMXRC3_UNEXPECTED_ERROR;
    }
  }
  return status;
}
