/*
** File: coredma.c  
** Project: ADB3 core driver
** Purpose: Implements functions for DMA transfers in Core Interface.
**
** (C) Copyright Alpha Data 2009-2010
*/

#include <df.h>

#include "resource.h"
#include "coredma.h"

#define DEBUGLEVEL_CANCEL (1) /* Level of debug messages relating to cancellation / cleanup */
#define DEBUGLEVEL_ERROR  (0) /* Level of debug messages relating to driver logic errors */
#define DEBUGLEVEL_NORMAL (6) /* Level of debug messages relating to normal operation */

typedef enum _MainState {
  MainIdle = 0,
  MainInitial = 1,
  MainRunning = 2
} MainState;

typedef enum _SetupState {
  SetupIdle = 0,
  SetupRunning = 1
} SetupState;

typedef enum _TransferState {
  TransferIdle = 0,
  TransferRunning = 1
} TransferState;

typedef enum _TeardownState {
  TeardownIdle = 0,
  TeardownRunning = 1
} TeardownState;

/* Forward declarations */

static void
dmaMapperCallback(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex,
  DfDmaMapperNode* pTable,
  unsigned int tableLength,
  DfDmaMapperPosition* pProgress,
  void* pContext);

/* Low-level functions */

/* Called while holding lock */
static void
dmaStartOnHardware(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channelIndex)
{
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  unsigned int tableIndex = pDmaContext->curr.state.transfer.tableIndex;

  /* TO DO - minimize the number of array accesses of 'pDmaContext->curr.table[]' and 'pDmaContext->descriptor[tableIndex]' */

  pDevCtx->methods.pDmaList(
    pDevCtx,
    pDmaContext->curr.parameters.bWriteToLocal,
    pDmaContext->curr.parameters.bFixedLocalAddress,
    pDmaContext->curr.table[tableIndex].pElements,
    pDmaContext->curr.table[tableIndex].length,
    &pDmaContext->curr.table[tableIndex].position,
    pDmaContext->descriptor[tableIndex].buffer.pGeneric,
    pDevCtx->dma.mapLength,
    pDmaContext->descriptor[tableIndex].busAddress);
  pDevCtx->methods.pDmaTransfer(
    pDevCtx,
    channelIndex,
    TRUE,
    pDmaContext->descriptor[tableIndex].buffer.pGeneric,
    pDmaContext->descriptor[tableIndex].busAddress);
}

static void
dmaStopOnHardware(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channelIndex)
{
  pDevCtx->methods.pDmaTransfer(pDevCtx, channelIndex, FALSE, NULL, 0);
}

static boolean_t
dmaSetup(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channelIndex)
{
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  unsigned int tableIndex = pDmaContext->curr.state.setup.tableIndex;

  return dfDmaMapperSetupThread(
    pDevCtx->pDevObj,
    (channelIndex << 1) | tableIndex,
    pDmaContext->curr.parameters.pBuffer,
    dmaMapperCallback,
    pDevCtx,
    pDmaContext->curr.parameters.bWriteToLocal,
    pDmaContext->curr.parameters.bFixedLocalAddress,
    &pDmaContext->curr.slice,
    1,
    &pDmaContext->curr.position);
}

static void
dmaTeardown(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channelIndex)
{
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  unsigned int tableIndex = pDmaContext->curr.state.teardown.tableIndex;

  pDmaContext->curr.state.teardown.tableIndex = tableIndex ^ 0x1U;
  dfDmaMapperTeardown(pDevCtx->pDevObj, (channelIndex << 1) | tableIndex);
}

static void
dmaCleanup(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channelIndex)
{
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];

  while (pDmaContext->curr.state.setup.count) {
    dmaTeardown(pDevCtx, channelIndex);
    pDmaContext->curr.state.setup.count--;
  }
  pDmaContext->curr.state.setup.count = 0;
  pDmaContext->curr.state.setup.tableIndex = 0;
  pDmaContext->curr.state.transfer.count = 0;
  pDmaContext->curr.state.transfer.tableIndex = 0;
  pDmaContext->curr.state.teardown.count = 0;
  pDmaContext->curr.state.teardown.tableIndex = 0;
}

static boolean_t
bCancelling(
  DmaChannelContext* pDmaContext)
{
  return pDmaContext->curr.state.bCancel;
}

static boolean_t
bCanCancel(
  DmaChannelContext* pDmaContext)
{
  return (pDmaContext->curr.state.setup.state == SetupIdle && pDmaContext->curr.state.transfer.state == TransferIdle && pDmaContext->curr.state.teardown.state == TeardownIdle) ? TRUE : FALSE;
}

static boolean_t
bCanMap(
  DmaChannelContext* pDmaContext)
{
  return (pDmaContext->curr.state.setup.state == SetupIdle && pDmaContext->curr.state.setup.count < 2 && pDmaContext->curr.position.sliceIndex != 1) ? TRUE : FALSE;
}

static boolean_t
bCanTransfer(
  DmaChannelContext* pDmaContext)
{
  return (pDmaContext->curr.state.transfer.state == TransferIdle && pDmaContext->curr.state.transfer.count > 0) ? TRUE : FALSE;
}

static boolean_t
bCanUnmap(
  DmaChannelContext* pDmaContext)
{
  return (pDmaContext->curr.state.teardown.state == TeardownIdle && pDmaContext->curr.state.teardown.count > 0) ? TRUE : FALSE;
}

static boolean_t
bTransferFinished(
  DmaChannelContext* pDmaContext)
{
  unsigned int tableIndex = pDmaContext->curr.state.transfer.tableIndex;

  return (pDmaContext->curr.table[tableIndex].position.index >= pDmaContext->curr.table[tableIndex].length) ? TRUE : FALSE;
}

static boolean_t
bOperationFinished(
  DmaChannelContext* pDmaContext)
{
  return (pDmaContext->curr.position.sliceIndex == 1 /* TO DO */ && pDmaContext->curr.state.setup.count == 0 && pDmaContext->curr.state.teardown.count == 0) ? TRUE : FALSE;
}

/* Callbacks */

static void
dmaMapperCallback(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex,
  DfDmaMapperNode* pTable,
  unsigned int tableLength,
  DfDmaMapperPosition* pProgress,
  void* pContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pContext;
  unsigned int channelIndex = mapIndex >> 1;
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  unsigned int tableIndex = pDmaContext->curr.state.setup.tableIndex;
  DfSpinLockFlags f;

  dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaMapperCallback: pDevCtx=%p mapIndex=%lu pTable=%p tableLength=%lu\n",
    (void*)pDevCtx, (unsigned long)mapIndex, (void*)pTable, (unsigned long)tableLength));

  pDmaContext->curr.table[tableIndex].pElements = pTable;
  pDmaContext->curr.table[tableIndex].length = tableLength;
  pDmaContext->curr.table[tableIndex].position.index = 0U;
  pDmaContext->curr.table[tableIndex].position.offset = 0U;
  pDmaContext->curr.state.setup.tableIndex = tableIndex ^ 0x1;
  pDmaContext->curr.position = *pProgress;
  f = dfSpinLockGet(&pDmaContext->lock);
  pDmaContext->curr.state.setup.state = SetupIdle;
  pDmaContext->curr.state.setup.count++;
  dfAssert(pDmaContext->curr.state.setup.count <= 2);
  pDmaContext->curr.state.transfer.count++;
  dfAssert(pDmaContext->curr.state.transfer.count <= 2);
  while (1) {
    if (bCancelling(pDmaContext)) {
      /* The DMA transfer is being cancelled */
      if (pDmaContext->curr.state.setup.state == SetupIdle && pDmaContext->curr.state.transfer.state == TransferIdle && pDmaContext->curr.state.teardown.state == TeardownIdle) {
        dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaMapperCallback: cancelled, invoking callback, channelIndex=%lu\n", (unsigned long)channelIndex));
        pDmaContext->curr.state.main = MainIdle;
        pDmaContext->curr.state.bCancel = FALSE;
        dfSpinLockPut(&pDmaContext->lock, f);
        dmaCleanup(pDevCtx, channelIndex);
        pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaCancelled);
        break;
      } else {
        dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaMapperCallback: cancelled, waiting for all idle, channelIndex=%lu\n", (unsigned long)channelIndex));
        dfSpinLockPut(&pDmaContext->lock, f);
        break;
      }
    }
    if (bCanTransfer(pDmaContext)) {
      /* Start DMA transfer on hardware */
      dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaMapperCallback: starting DMA transfer on hardware, channelIndex=%lu\n", (unsigned long)channelIndex));
      pDmaContext->curr.state.transfer.state = TransferRunning;
      dmaStartOnHardware(pDevCtx, channelIndex);
    } else {
      if (bCanMap(pDmaContext)) {
        /* Set up DMA transfer */
        dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaMapperCallback: setting up DMA transfer, channelIndex=%lu\n", (unsigned long)channelIndex));
        pDmaContext->curr.state.setup.state = SetupRunning;
        dfSpinLockPut(&pDmaContext->lock, f);
        dmaSetup(pDevCtx, channelIndex);
        f = dfSpinLockGet(&pDmaContext->lock);
      } else {
        if (bCanUnmap(pDmaContext)) {
          /* Tear down DMA transfer */
          dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaMapperCallback: tearing down DMA transfer, channelIndex=%lu\n", (unsigned long)channelIndex));
          pDmaContext->curr.state.teardown.state = TeardownRunning;
          dfSpinLockPut(&pDmaContext->lock, f);
          dmaTeardown(pDevCtx, channelIndex);
          f = dfSpinLockGet(&pDmaContext->lock);
          pDmaContext->curr.state.setup.count--;
          dfAssert(pDmaContext->curr.state.setup.count <= 2);
          pDmaContext->curr.state.teardown.count--;
          dfAssert(pDmaContext->curr.state.teardown.count <= 2);
          pDmaContext->curr.state.teardown.state = TeardownIdle;
          if (bOperationFinished(pDmaContext)) {
            /* The DMA transfer is finished */
            dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaMapperCallback: DMA transfer finished, channelIndex=%lu\n", (unsigned long)channelIndex));
            dfAssert(pDmaContext->curr.state.transfer.count == 0);
            pDmaContext->curr.state.main = MainIdle;
            dfSpinLockPut(&pDmaContext->lock, f);
            pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaSuccess);
            break;
          }
        } else {
          /* Nothing more to do */
          dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaMapperCallback: nothing more to do, channelIndex=%lu\n", (unsigned long)channelIndex));
          dfSpinLockPut(&pDmaContext->lock, f);
          break;
        }
      }
    }
  }
}

void
dmaCompletedRoutine(
  DfDpc* pDpc,
  void* pContext, /* channel */
  void* pArg)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pContext;
  unsigned int channelIndex = (unsigned int)(uintptr_t)pArg;
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  DfSpinLockFlags f;

  dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaCompletedRoutine: pDevCtx=%p channelIndex=%lu\n", (void*)pDevCtx, (unsigned long)channelIndex));

  f = dfSpinLockGet(&pDmaContext->lock);
  if (pDmaContext->curr.state.transfer.state == TransferIdle) {
    dfSpinLockPut(&pDmaContext->lock, f);
    dfDebugPrint(DEBUGLEVEL_ERROR, ("+++ dmaCompletedRoutine: spurious call, pDevCtx=%p channelIndex=%lu\n", (void*)pDevCtx, (unsigned long)channelIndex));
    return;
  }
  pDmaContext->curr.state.transfer.state = TransferIdle;
  if (bTransferFinished(pDmaContext)) {
    pDmaContext->curr.state.transfer.tableIndex = pDmaContext->curr.state.transfer.tableIndex ^ 0x1;
    pDmaContext->curr.state.transfer.count--;
    dfAssert(pDmaContext->curr.state.transfer.count <= 2);
    pDmaContext->curr.state.teardown.count++;
    dfAssert(pDmaContext->curr.state.teardown.count <= 2);
  }
  while (1) {
    if (bCancelling(pDmaContext)) {
      /* The DMA transfer is being cancelled */
      if (bCanCancel(pDmaContext)) {
        dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaCompletedRoutine: cancelled, invoking callback, channelIndex=%lu\n", (unsigned long)channelIndex));
        pDmaContext->curr.state.main = MainIdle;
        pDmaContext->curr.state.bCancel = FALSE;
        dfSpinLockPut(&pDmaContext->lock, f);
        dmaCleanup(pDevCtx, channelIndex);
        pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaCancelled);
        break;
      } else {
        dfDebugPrint(DEBUGLEVEL_CANCEL, ("dmaCompletedRoutine: cancelled, waiting for all idle, channelIndex=%lu\n", (unsigned long)channelIndex));
        dfSpinLockPut(&pDmaContext->lock, f);
        break;
      }
    }
    if (bCanTransfer(pDmaContext)) {
      /* Start DMA transfer on hardware */
      dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaCompletedRoutine: starting DMA transfer on hardware, channelIndex=%lu\n", (unsigned long)channelIndex));
      pDmaContext->curr.state.transfer.state = TransferRunning;
      dmaStartOnHardware(pDevCtx, channelIndex);
    } else {
      if (bCanMap(pDmaContext)) {
        /* Setup DMA transfer */
        dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaCompletedRoutine: setting up DMA transfer, channelIndex=%lu\n", (unsigned long)channelIndex));
        pDmaContext->curr.state.setup.state = SetupRunning;
        dfSpinLockPut(&pDmaContext->lock, f);
        dmaSetup(pDevCtx, channelIndex);
        f = dfSpinLockGet(&pDmaContext->lock);
      } else {
        if (bCanUnmap(pDmaContext)) {
          /* Tear down DMA transfer */
          dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaCompletedRoutine: tearing down DMA transfer, channelIndex=%lu\n", (unsigned long)channelIndex));
          pDmaContext->curr.state.teardown.state = TeardownRunning;
          dfSpinLockPut(&pDmaContext->lock, f);
          dmaTeardown(pDevCtx, channelIndex);
          f = dfSpinLockGet(&pDmaContext->lock);
          pDmaContext->curr.state.setup.count--;
          dfAssert(pDmaContext->curr.state.setup.count <= 2);
          pDmaContext->curr.state.teardown.count--;
          dfAssert(pDmaContext->curr.state.teardown.count <= 2);
          pDmaContext->curr.state.teardown.state = TeardownIdle;
          if (bOperationFinished(pDmaContext)) {
            /* The DMA transfer is finished */
            dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaCompletedRoutine: DMA transfer finished, channelIndex=%lu\n", (unsigned long)channelIndex));
            dfAssert(pDmaContext->curr.state.transfer.count == 0);
            pDmaContext->curr.state.main = MainIdle;
            dfSpinLockPut(&pDmaContext->lock, f);
            pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaSuccess);
            break;
          }
        } else {
          /* Nothing more to do */
          dfDebugPrint(DEBUGLEVEL_NORMAL, ("dmaCompletedRoutine: nothing more to do, channelIndex=%lu\n", (unsigned long)channelIndex));
          dfSpinLockPut(&pDmaContext->lock, f);
          break;
        }
      }
    }
  }
}

/*
** Handle CoreDmaCancel
*/

CoreDmaStatus
coreDmaCancel(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int channelIndex)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  DfSpinLockFlags f;

  if (channelIndex >= pDevCtx->info.bootstrap.numDmaChannel) {
    return CoreDmaInvalidChannel;
  }

  if (!checkTicketValid(pDevCtx, pTicket, CORE_RESOURCE_DMAENGINE, channelIndex)) {
    return CoreDmaInvalidTicket;
  }

  f = dfSpinLockGet(&pDmaContext->lock);
  pDmaContext->curr.state.bCancel = TRUE;
  switch (pDmaContext->curr.state.main) {
  case MainInitial:
    dfDebugPrint(DEBUGLEVEL_CANCEL, ("coreDmaCancel: in Initial state\n"));
    dfSpinLockPut(&pDmaContext->lock, f);
    break;

  case MainRunning:
    if (pDmaContext->curr.state.setup.state == SetupIdle && pDmaContext->curr.state.transfer.state == TransferIdle && pDmaContext->curr.state.teardown.state == TeardownIdle) {
      dfDebugPrint(DEBUGLEVEL_CANCEL, ("coreDmaCancel: in Running state, cleaning up\n"));
      pDmaContext->curr.state.main = MainIdle;
      dfSpinLockPut(&pDmaContext->lock, f);
      dmaCleanup(pDevCtx, channelIndex);
      pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaCancelled);
    } else {
      if (pDmaContext->curr.state.transfer.state == TransferRunning) {
        dfDebugPrint(DEBUGLEVEL_CANCEL, ("coreDmaCancel: in Running state, aborting DMA on hardware\n"));
        dmaStopOnHardware(pDevCtx, channelIndex);
      } else {
        dfDebugPrint(DEBUGLEVEL_CANCEL, ("coreDmaCancel: in Running state, not aborting DMA on hardware\n"));
      }
      dfSpinLockPut(&pDmaContext->lock, f);
    }
    break;

  default:
    dfSpinLockPut(&pDmaContext->lock, f);
    dfDebugPrint(DEBUGLEVEL_ERROR, ("*** coreDmaCancel: in INVALID state (%u)\n", pDmaContext->curr.state.main));
    break;
  }

  return TRUE;
}

/*
** Handle CoreDmaInitialize
*/

CoreDmaStatus
coreDmaInitialize(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int channelIndex,
  const CoreDmaParameters* pParameters,
  CoreDmaCallback* pCallback,
  void* pCallbackContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];

  if (channelIndex >= pDevCtx->info.bootstrap.numDmaChannel) {
    return CoreDmaInvalidChannel;
  }

  if (!checkTicketValid(pDevCtx, pTicket, CORE_RESOURCE_DMAENGINE, channelIndex)) {
    return CoreDmaInvalidTicket;
  }

  dfAssert(pDmaContext->curr.state.main != MainRunning);

  pDmaContext->curr.pTicket = pTicket;
  pDmaContext->curr.parameters = *pParameters;
  pDmaContext->curr.pCallback = pCallback;
  pDmaContext->curr.pCallbackContext = pCallbackContext;
  pDmaContext->curr.slice.length = pParameters->region.length;
  pDmaContext->curr.slice.offset = pParameters->region.offset;
  pDmaContext->curr.slice.localAddress = pParameters->region.localAddress;
  pDmaContext->curr.position.sliceIndex = 0;
  pDmaContext->curr.position.slicePosition = 0;

  pDmaContext->curr.state.bCancel = FALSE;
  pDmaContext->curr.state.main = MainInitial;
  pDmaContext->curr.state.setup.tableIndex = 0;
  pDmaContext->curr.state.transfer.tableIndex = 0;
  pDmaContext->curr.state.teardown.tableIndex = 0;

  return CoreDmaSuccess;
}

/*
** Handle CoreDmaTransfer
*/

void
coreDmaStart(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int channelIndex)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;
  DmaChannelContext* pDmaContext = &pDevCtx->dma.channel[channelIndex];
  DfSpinLockFlags f;

  if (channelIndex >= pDevCtx->info.bootstrap.numDmaChannel) {
    pDmaContext->curr.pCallback(pTicket, pDmaContext->curr.pCallbackContext, CoreDmaInvalidChannel);
  }

  if (!checkTicketValid(pDevCtx, pTicket, CORE_RESOURCE_DMAENGINE, channelIndex)) {
    pDmaContext->curr.pCallback(pTicket, pDmaContext->curr.pCallbackContext, CoreDmaInvalidTicket);
  }

  f = dfSpinLockGet(&pDmaContext->lock);
  dfAssert(pDmaContext->curr.state.main == MainInitial);
  dfAssert(pDmaContext->curr.state.setup.state == SetupIdle);
  dfAssert(pDmaContext->curr.state.setup.count == 0);
  dfAssert(pDmaContext->curr.state.transfer.state == TransferIdle);
  dfAssert(pDmaContext->curr.state.teardown.state == TeardownIdle);
  dfAssert(pDmaContext->curr.state.teardown.count == 0);
  if (pDmaContext->curr.state.bCancel) {
    /* We were cancelled */
    dfDebugPrint(DEBUGLEVEL_CANCEL, ("coreDmaStart: cancelled, channelIndex=%lu\n", (unsigned long)channelIndex));
    pDmaContext->curr.state.bCancel = FALSE;
    pDmaContext->curr.state.main = MainIdle;
    dfSpinLockPut(&pDmaContext->lock, f);
    pDmaContext->curr.pCallback(pTicket, pDmaContext->curr.pCallbackContext, CoreDmaCancelled);
  } else {
    dfDebugPrint(DEBUGLEVEL_NORMAL, ("coreDmaStart: starting setup, channelIndex=%lu\n", (unsigned long)channelIndex));
    pDmaContext->curr.state.main = MainRunning;
    pDmaContext->curr.state.setup.state = SetupRunning;
    dfSpinLockPut(&pDmaContext->lock, f);
    if (!dmaSetup(pDevCtx, channelIndex)) {
      f = dfSpinLockGet(&pDmaContext->lock);
      pDmaContext->curr.state.setup.state = SetupIdle;
      pDmaContext->curr.state.main = MainIdle;
      if (pDmaContext->curr.state.bCancel) {
        dfDebugPrint(DEBUGLEVEL_CANCEL, ("coreDmaStart: setup failed, cancelled, channelIndex=%lu\n", (unsigned long)channelIndex));
        dfSpinLockPut(&pDmaContext->lock, f);
        pDmaContext->curr.pCallback(pDmaContext->curr.pTicket, pDmaContext->curr.pCallbackContext, CoreDmaCancelled);
      } else {
        dfDebugPrint(DEBUGLEVEL_CANCEL, ("coreDmaStart: setup failed, channelIndex=%lu\n", (unsigned long)channelIndex));
        dfSpinLockPut(&pDmaContext->lock, f);
        pDmaContext->curr.pCallback(pTicket, pDmaContext->curr.pCallbackContext, CoreDmaUnexpectedError);
      }
    }
  }
}

unsigned int
coreGetNumDmaChannel(
  void* pInterfaceContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;

  return pDevCtx->info.bootstrap.numDmaChannel;
}
