/*
** File: model_adb3.c  
** Project: ADB3 core driver
** Purpose: ADB3 generic code, including function to bootstrap device context.
**
** (C) Copyright Alpha Data 2009-2010
*/

#include <df.h>

#include "adb3.h"
#include "adb3_common.h"
#include "device.h"
#include "model_boot.h"
#include "flash.h"

boolean_t
adb3VerifyChecksum(
	Adb3CoreDeviceContext* pDevCtx,
  uint16_t vpdLength)
{
  uint8_t buff[16];
  uint8_t sum = 0, checksum;
  uint16_t offset = 0;
  uint16_t remaining = vpdLength - 1;
  uint16_t i, chunk;

  if (vpdLength < 5) {
  	dfDebugPrint(0, ("*** verifyChecksum: invalid VPD length %lu(0x%lx)\n",
      (unsigned long)vpdLength, (unsigned long)vpdLength));
    pDevCtx->info.vpd.bChecksumError = TRUE;
    return FALSE;
  }

  /* Get checksum from VPD memory; always last byte of structure */
  flashRead(pDevCtx, 0, pDevCtx->info.bootstrap.vpdOffset + remaining, sizeof(checksum), &checksum);

  /* Compute checksum over all but last byte of structure */
  while (remaining) {
    chunk = (remaining > sizeof(buff)) ? sizeof(buff) : remaining;
    flashRead(pDevCtx, 0, pDevCtx->info.bootstrap.vpdOffset + offset, chunk, buff);
    for (i = 0; i < chunk; i++) {
      sum = sum + buff[i];
    }
    offset = offset + chunk;
    remaining = remaining - chunk;
  }
  sum = (uint8_t)-sum;

  if (sum == checksum) {
    return TRUE;
  } else {
  	dfDebugPrint(0, ("*** verifyChecksum: checksum invalid, computed 0x%02lx found 0x%02lx\n",
      (unsigned long)sum, (unsigned long)checksum));
    pDevCtx->info.vpd.bChecksumError = TRUE;
    return FALSE;
  }
}

void
adb3DmaList(
  Adb3CoreDeviceContext* pDevCtx,
  boolean_t bWriteToDevice,
  boolean_t bFixedLocal,
  DfDmaMapperNode* pTable,
  unsigned int tableLength,
  DmaNodeTablePosition* pTablePosition,
  void* pDescriptorBuffer,
  unsigned int maxDescriptor,
  uint64_t descriptorBusAddress)
{
  const boolean_t bWorkAround4kCrossing = pDevCtx->info.bootstrap.bWorkAround4kCrossing;
  const unsigned int fourk = 0x1000U, fourkMask = fourk - 1U;
  Adb3DmaDescriptor* pDescriptor = (Adb3DmaDescriptor*)pDescriptorBuffer;
  DfDmaMapperNode* pTableTmp;
  uint64_t busAddress, nextBusAddress, tableBusAddress = 0U, localAddress = 0U;
  unsigned int i;
  uint32_t ctl;
  size_t remaining;
  unsigned int chunk, maxChunk;
  unsigned int numDesc = 0;
  unsigned int index;
  size_t offset;

  dfDebugPrint(7, ("adb3DmaList: entered, pDevCtx=%p bWriteToDevice=%s bFixedLocal=%s pTable=%p tableLength=%u *pTablePosition={%u, 0x%llx} pDescriptorBuffer=%p maxDescriptor=%u descriptorBusAddrss=0x%llx\n",
    (void*)pDevCtx, bWriteToDevice ? "TRUE" : "FALSE", bFixedLocal ? "TRUE" : "FALSE", (void*)pTable, tableLength, pTablePosition->index, (unsigned long long)pTablePosition->offset, pDescriptorBuffer, maxDescriptor, (unsigned long long)descriptorBusAddress));

  ctl = bWriteToDevice ? ADB3DMACTL_PCI_TO_LOCAL : ADB3DMACTL_LOCAL_TO_PCI;
  ctl |= bFixedLocal ? ADB3DMACTL_FIXED_LOCAL_ADDR : 0;
  busAddress = descriptorBusAddress;
  index = pTablePosition->index;
  offset = pTablePosition->offset;
  pTableTmp = pTable + pTablePosition->index;
  if (bWorkAround4kCrossing) {
    i = index;
    remaining = 0;
    for (i = index; i < tableLength && numDesc < maxDescriptor; numDesc++) {
      if (0 == remaining) {
        tableBusAddress = pTableTmp->busAddress + offset;
        if (!bFixedLocal) {
          localAddress = pTableTmp->localAddress + offset;
        }
        remaining = pTableTmp->length - offset;
        maxChunk = fourk - ((unsigned int)tableBusAddress & fourkMask);
      } else {
        maxChunk = fourk;
      }
      if (maxChunk > remaining) {
        maxChunk = (unsigned int)remaining; /* Cast is safe due to above comparison */
      }
      nextBusAddress = busAddress + sizeof(Adb3DmaDescriptor);
      /* Cast is safe since ((unsigned int)remaining) only used if remaining <= maxChunk */
      chunk = (remaining > maxChunk) ? maxChunk : (unsigned int)remaining;
      pDescriptor->pal = dfCpuToLe32((uint32_t)tableBusAddress);
      pDescriptor->pah = dfCpuToLe32((uint32_t)(tableBusAddress >> 32));
      pDescriptor->sizl = dfCpuToLe32((uint32_t)chunk);
      pDescriptor->sizh = dfCpuToLe32(0U);
      pDescriptor->ndl = dfCpuToLe32((uint32_t)nextBusAddress);
      pDescriptor->ndh = dfCpuToLe32((uint32_t)(nextBusAddress >> 32));
      pDescriptor->lal = dfCpuToLe32((uint32_t)localAddress);
      pDescriptor->ctl = dfCpuToLe32(ctl | ((uint32_t)(localAddress >> 32) & 0x7FU));
      pDescriptor++;
      busAddress = nextBusAddress;
      tableBusAddress += chunk;
      if (!bFixedLocal) {
        localAddress += chunk;
      }
      offset += chunk;
      remaining -= chunk;
      if (0 == remaining) {
        pTableTmp++;
        i++;
        offset = 0;
      }
    }
  } else {
    for (i = index; i < tableLength && numDesc < maxDescriptor; i++, numDesc++) {
      nextBusAddress = busAddress + sizeof(Adb3DmaDescriptor);
      tableBusAddress = pTableTmp->busAddress + offset;
      localAddress = pTableTmp->localAddress;
      if (!bFixedLocal) {
        localAddress += offset;
      }
      remaining = pTableTmp->length - offset;
      pDescriptor->pal = dfCpuToLe32((uint32_t)tableBusAddress);
      pDescriptor->pah = dfCpuToLe32((uint32_t)(tableBusAddress >> 32));
      pDescriptor->sizl = dfCpuToLe32((uint32_t)remaining);
      pDescriptor->sizh = dfCpuToLe32(0U);
      pDescriptor->ndl = dfCpuToLe32((uint32_t)nextBusAddress);
      pDescriptor->ndh = dfCpuToLe32((uint32_t)(nextBusAddress >> 32));
      pDescriptor->lal = dfCpuToLe32((uint32_t)localAddress);
      pDescriptor->ctl = dfCpuToLe32(ctl | ((uint32_t)(localAddress >> 32) & 0x7FU));
      pTableTmp++;
      busAddress = nextBusAddress;
      offset = 0;
      pDescriptor++;
    }
  }
  pDescriptor--;
  pDescriptor->ndl = dfCpuToLe32(0U);
  pDescriptor->ctl |= dfCpuToLe32(ADB3DMACTL_ENDCHAIN | (bWriteToDevice ? ADB3DMACTL_IRQ_ON_FPGA_COMPLETION : ADB3DMACTL_IRQ_ON_HOST_COMPLETION));

  /* Cast is safe as pTableTmp cannot be much greater than pTable */
  pTablePosition->index = (unsigned int)(pTableTmp - pTable);
  pTablePosition->offset = offset;
  dfDebugPrint(7, ("adb3DmaList: finished, *pTablePosition={%u, 0x%llx}\n", pTablePosition->index, (unsigned long long)pTablePosition->offset));

#if DF_DBG_BUILD
  if (g_dfDebugLevel >= 7) {
    dfDebugPrint(0, ("adb3DmaList: dump of DMA engine linked list\n"));
    adb3DumpDmaDescriptors(pDevCtx, (Adb3DmaDescriptor*)pDescriptorBuffer, descriptorBusAddress, numDesc);
  }
#endif
}

void
adb3DmaTransfer(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channel,
  boolean_t bStart,
  void* pDescriptorBuffer,
  uint64_t descriptorBusAddress)
{
  const unsigned int maxAbortPoll = 20;
  Adb3DmaDescriptor* pDescriptor = (Adb3DmaDescriptor*)pDescriptorBuffer;
  Adb3GenericRegs* pRegs = pDevCtx->hardware.bridge.pAdb3;
  Adb3DmaRegs* pDmaRegs = &pRegs->dmaEngine[channel];
  uint32_t alignment;
  uint32_t lal, pal, ctl, status;
  DfMemoryHandle handle;
  unsigned int i;

  dfDebugPrint(6, ("adb3DmaTransfer: entered, pDevCtx=%p channel=%lu bStart=%s pDescriptorBuffer=%p busAddress=0x%08lx_%08lx\n",
    (void*)pDevCtx, (unsigned long)channel, bStart ? "TRUE" : "FALSE", pDescriptorBuffer, dfSplitUint64(descriptorBusAddress)));

  handle = pDevCtx->hardware.hBridge;

  if (bStart) {
    lal = dfLe32ToCpu(pDescriptor->lal & 0xfU);
    pal = dfLe32ToCpu(pDescriptor->pal & 0xfU);
    ctl = dfLe32ToCpu(pDescriptor->ctl);
    if (ctl & ADB3DMACTL_PCI_TO_LOCAL) {
      if (pal > lal) {
        alignment = ((0x10U + lal - pal) & 0xfU) | ADB3DMAALIGN_LOCAL_IGNORE_FIRST_WORD;
      } else {
        alignment = (lal - pal) | ADB3DMAALIGN_LOCAL_USE_FIRST_WORD;
      }
      dfPciMemWrite32(handle, &pDmaRegs->irqEnable, ADB3DMAIRQ_MPTL_WRITE_WORKER_COMPLETED | ADB3DMAIRQ_ABORT);
    } else {
      if (pal >= lal) {
        alignment = ((0x10U + lal - pal) & 0xf) | ADB3DMAALIGN_LOCAL_USE_FIRST_WORD;
      } else {
        alignment = (lal - pal) | ADB3DMAALIGN_LOCAL_IGNORE_FIRST_WORD;
      }
      dfPciMemWrite32(handle, &pDmaRegs->irqEnable, ADB3DMAIRQ_PCIE_WRITE_WORKER_COMPLETED | ADB3DMAIRQ_ABORT);
    }
    dfPciMemWrite32(handle, &pDmaRegs->alignment, alignment);
    dfPciMemWrite32(handle, &pDmaRegs->nextAddrHigh, (uint32_t)(descriptorBusAddress >> 32));

#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("adb3DmaTransfer: DMA channel %lu registers before starting on hardware:\n", (unsigned long)channel));
      adb3DumpDmaRegisters(pDevCtx, channel);
    }
#endif

    /* Writing to nextAddrLow starts the DMA transfer */
    dfPciMemWrite32(handle, &pDmaRegs->nextAddrLow, (uint32_t)descriptorBusAddress); 

#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("adb3DmaTransfer: DMA channel %lu registers after starting on hardware:\n", (unsigned long)channel));
      adb3DumpDmaRegisters(pDevCtx, channel);
    }
#endif
  } else {
#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("adb3DmaTransfer: DMA channel %lu registers before attempt to abort:\n", (unsigned long)channel));
      adb3DumpDmaRegisters(pDevCtx, channel);
    }
#endif

    /* Write to ABORT register */ 
    dfPciMemWrite32(handle, &pDmaRegs->control.abort, 0);
    for (i = 0; i < maxAbortPoll; i++) {
      status = dfPciMemRead32(handle, &pDmaRegs->control.status);
    }
    /* Assert cleanup strobe - should generate an interrupt */
    dfPciMemWrite32(handle, &pDmaRegs->cleanup.cleanup, 0);
    dfPciMemRead32(handle, &pDmaRegs->cleanup.cleanup);

#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("adb3DmaTransfer: DMA channel %lu registers after attempt to abort:\n", (unsigned long)channel));
      adb3DumpDmaRegisters(pDevCtx, channel);
    }
#endif
  }
}

void
adb3DumpDmaDescriptors(
	Adb3CoreDeviceContext* pDevCtx,
  Adb3DmaDescriptor* pDescriptors,
  uint64_t descBus,
  unsigned int count)
{
  Adb3DmaDescriptor* pCurrent;
  unsigned int i;

  if (sizeof(void*) == 4) {
    dfDebugPrint(0, ("  ### DescVirt   DescBus              [ PAH      PAL      SIZH     SIZL     NDH      NDL      LAL      CTL      ]\n"));
    for (i = 0; i < count; i++) {
      pCurrent = &pDescriptors[i];
      dfDebugPrint(0, ("  %03lu %08p 0x%08lx_%08lx   %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
        (unsigned long)i, (void*)pCurrent, dfSplitUint64(descBus),
        (unsigned long)dfLe32ToCpu(pCurrent->pah), (unsigned long)dfLe32ToCpu(pCurrent->pal),
        (unsigned long)dfLe32ToCpu(pCurrent->sizh), (unsigned long)dfLe32ToCpu(pCurrent->sizl),
        (unsigned long)dfLe32ToCpu(pCurrent->ndh), (unsigned long)dfLe32ToCpu(pCurrent->ndl),
        (unsigned long)dfLe32ToCpu(pCurrent->lal), (unsigned long)dfLe32ToCpu(pCurrent->ctl)));
      descBus += sizeof(Adb3DmaDescriptor);
    }
  } else {
    dfDebugPrint(0, ("  ### DescVirt           DescBus              [ PAH      PAL      SIZH     SIZL     NDH      NDL      LAL      CTL      ]\n"));
    for (i = 0; i < count; i++) {
      pCurrent = &pDescriptors[i];
      dfDebugPrint(0, ("  %03lu %016p 0x%08lx_%08lx   %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
        (unsigned long)i, (void*)pCurrent, dfSplitUint64(descBus),
        (unsigned long)dfLe32ToCpu(pCurrent->pah), (unsigned long)dfLe32ToCpu(pCurrent->pal),
        (unsigned long)dfLe32ToCpu(pCurrent->sizh), (unsigned long)dfLe32ToCpu(pCurrent->sizl),
        (unsigned long)dfLe32ToCpu(pCurrent->ndh), (unsigned long)dfLe32ToCpu(pCurrent->ndl),
        (unsigned long)dfLe32ToCpu(pCurrent->lal), (unsigned long)dfLe32ToCpu(pCurrent->ctl)));
      descBus += sizeof(Adb3DmaDescriptor);
    }
  }
}

void
adb3DumpDmaRegisters(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int channel)
{
  static const char* regNames[16] = {
    "status", "fifoStatus", "irqStatus", "irqEnable",
    "nextAddrLow", "nextAddrHigh", "hostControl", "fpgaControl",
    "alignment", "descPause", "reservedA", "reservedB",
    "reservedC", "reservedD", "reservedE", "reservedF"
  };
  DfMemoryHandle handle;
  uint32_t* pRegs = (uint32_t*)&pDevCtx->hardware.bridge.pAdb3->dmaEngine[channel];
  uint32_t regValue[16];
  unsigned int i;

  handle = pDevCtx->hardware.hBridge;
  for (i = 0; i < 16; i++) {
    regValue[i] = dfPciMemRead32(handle, pRegs + i);
  }
  for (i = 0; i < 4; i++) {
    dfDebugPrint(0, ("  %12s=0x%08lx %12s=0x%08lx %12s=0x%08lx %12s=0x%08lx\n",
      regNames[i], (unsigned long)regValue[i],
      regNames[i + 4], (unsigned long)regValue[i + 4],
      regNames[i + 8], (unsigned long)regValue[i + 8],
      regNames[i + 12], (unsigned long)regValue[i + 12]));
  }
}

static boolean_t
initHardware(
	Adb3CoreDeviceContext* pDevCtx,
	boolean_t bInit)
{
  Adb3GenericRegs* pRegs;

  dfDebugPrint(1, ("initHardware: bInit=%s\n", bInit ? "TRUE" : "FALSE"));

  pRegs = pDevCtx->hardware.bridge.pAdb3;

  if (bInit) {
    pDevCtx->info.model = CoreModelGeneric;

    /* Find base addresses and sizes of local bus windows */
    pDevCtx->info.window[0].busBase = pDevCtx->rawResources.variant.pci.memBar[2].base;
    pDevCtx->info.window[0].busSize = pDevCtx->rawResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].localBase = 0;
    pDevCtx->info.window[0].localSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[2].base;
    pDevCtx->info.window[0].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].pKernelBase = pDevCtx->hardware.pBridged;
    pDevCtx->info.window[0].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[1].busBase = pDevCtx->rawResources.variant.pci.memBar[3].base;
    pDevCtx->info.window[1].busSize = pDevCtx->rawResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].localBase = 0;
    pDevCtx->info.window[1].localSize = pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[3].base;
    pDevCtx->info.window[1].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].pKernelBase = pDevCtx->hardware.pBridgedPrefetch;
    pDevCtx->info.window[1].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[2].busBase = pDevCtx->rawResources.variant.pci.memBar[0].base;
    pDevCtx->info.window[2].busSize = pDevCtx->rawResources.variant.pci.memBar[0].size;
    pDevCtx->info.window[2].localBase = 0;
    pDevCtx->info.window[2].localSize = 0;
    pDevCtx->info.window[2].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[0].base;
    pDevCtx->info.window[2].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[0].size;
    pDevCtx->info.window[2].pKernelBase = pDevCtx->hardware.bridge.pGeneric;
    pDevCtx->info.window[2].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[3].busBase = pDevCtx->rawResources.variant.pci.memBar[1].base;
    pDevCtx->info.window[3].busSize = pDevCtx->rawResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[3].localBase = 0;
    pDevCtx->info.window[3].localSize = 0;
    pDevCtx->info.window[3].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[1].base;
    pDevCtx->info.window[3].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[3].pKernelBase = pDevCtx->hardware.model.pGeneric;
    pDevCtx->info.window[3].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[1].size;
  } else {
    /* Nothing to do */
  }

  return TRUE;
}

static void
enableInterrupts(
	Adb3CoreDeviceContext* pDevCtx,
	boolean_t bEnable)
{
	Adb3GenericRegs* pRegs;
  
  pRegs = pDevCtx->hardware.bridge.pAdb3;

  dfDebugPrint(1, ("enableInterrupts: entered bEnable=%s\n", bEnable ? "TRUE" : "FALSE"));

	if (bEnable) {
		dfDebugPrint(1, ("enableInterrupts: interrupts enabled\n"));
	} else {
		dfDebugPrint(1, ("enableInterrupts: interrupts disabled\n"));
	}
}

static boolean_t
isr(
  DfInterruptObject* pInterruptObj,
  void* pInterruptContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterruptContext;
  boolean_t bHandled = FALSE;
  Adb3GenericRegs* pRegs;

  dfDebugPrint(1, ("isr: entered\n"));

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  
  return bHandled;
}

static boolean_t
mapBusResources(
	Adb3CoreDeviceContext* pDevCtx,
	DfBusResources* pRaw,
	DfBusResources* pTranslated,
	boolean_t bUnmap)
{
	dfDebugPrint(2, ("mapBusResources: bUnmap=%s\n", bUnmap ? "TRUE" : "FALSE"));

	if (bUnmap) {
		if (NULL != pDevCtx->hardware.bridge.pGeneric) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.bridge.pGeneric)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 generic registers\n"));
			}
      pDevCtx->hardware.bridge.pGeneric = NULL;
		}
		if (NULL != pDevCtx->hardware.model.pGeneric) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.model.pGeneric)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 model-specific registers\n"));
			}
      pDevCtx->hardware.model.pGeneric = NULL;
		}
    if (NULL != pDevCtx->hardware.pBridged) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.pBridged)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 bridged space\n"));
			}
      pDevCtx->hardware.pBridged = NULL;
		}
    if (NULL != pDevCtx->hardware.pBridgedPrefetch) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.pBridgedPrefetch)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 prefetchable bridged space\n"));
			}
      pDevCtx->hardware.pBridgedPrefetch = NULL;
		}
		dfDebugPrint(2, ("mapBusResources: unmapped\n"));
	} else {
    pDevCtx->hardware.bridge.pGeneric = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 0, 0, 0x1000U, &pDevCtx->hardware.hBridge);
		if (NULL == pDevCtx->hardware.bridge.pGeneric) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 generic registers\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 generic registers @ 0x%08lx_%08lx to %p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[0].base), pDevCtx->hardware.bridge.pGeneric));
		}
		pDevCtx->hardware.model.pGeneric = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 1, 0, 0x1000U, &pDevCtx->hardware.hModel);
		if (NULL == pDevCtx->hardware.model.pGeneric) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 model-specific registers\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 model-specific registers @ 0x%08lx_%08lx to %p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[1].base), pDevCtx->hardware.model.pGeneric));
		}
    pDevCtx->hardware.pBridged = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 2, 0, pTranslated->variant.pci.memBar[2].size, &pDevCtx->hardware.hBridged);
		if (NULL == pDevCtx->hardware.pBridged) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 bridged space\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 bridged space @ 0x%08lx_%08lx to %p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[2].base), pDevCtx->hardware.pBridged));
		}
    pDevCtx->hardware.pBridgedPrefetch = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 3, 0, pTranslated->variant.pci.memBar[3].size, &pDevCtx->hardware.hBridgedPrefetch);
		if (NULL == pDevCtx->hardware.pBridgedPrefetch) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 prefetchable bridged space\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 prefetchable bridged space @ 0x%08lx_%08lx to %p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[3].base), pDevCtx->hardware.pBridgedPrefetch));
		}
		dfDebugPrint(2, ("mapBusResources: mapped OK\n"));
	}

	return TRUE;
}

boolean_t
validateBusResources(
	Adb3CoreDeviceContext* pDevCtx,
	DfBusResources* pRaw,
	DfBusResources* pTranslated)
{
	if (pTranslated->variant.pci.numMemBar < 4) {
		dfDebugPrint(0, ("*** validateBusResources: less than 4 memory BARs\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[0].size < 0x1000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR0 smaller than 0x1000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[1].size < 0x1000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR1 smaller than 0x1000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[2].size < 0x100000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR2 smaller than 0x100000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[3].size < 0x100000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR2 smaller than 0x100000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.numIrq != 1) {
		dfDebugPrint(0, ("*** validateBusResources: not exactly 1 interrupt\n"));
		return FALSE;
	}

	dfDebugPrint(2, ("validateBusResources: OK\n"));

	return TRUE;
}

void
adb3CoreBootstrapGenericAdb3(
	Adb3CoreDeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("adb3CoreBootstrapGenericAdb3: entered\n"));

  pDevCtx->methods.pEnableInterrupts = enableInterrupts;
  pDevCtx->methods.pInitHardware = initHardware;
  pDevCtx->methods.pIsr = isr;
  pDevCtx->methods.pMapBusResources = mapBusResources;
  pDevCtx->methods.pValidateBusResources = validateBusResources;

  pDevCtx->info.bootstrap.numClockGen = 0;
  pDevCtx->info.bootstrap.numDmaChannel = 1;
  pDevCtx->info.bootstrap.numIoModule = 1;
  pDevCtx->info.bootstrap.numSensor = 0;
  pDevCtx->info.bootstrap.numTargetFpga = 0;
  pDevCtx->info.bootstrap.numWindow = 4;
  pDevCtx->info.bootstrap.bVpdInFlash = TRUE;
  pDevCtx->info.bootstrap.vpdLength = 0;
  pDevCtx->info.bootstrap.vpdOffset = 0;
  pDevCtx->info.bootstrap.flashDeviceWidth = 0;
  pDevCtx->info.bootstrap.flashWidth = 0;
  pDevCtx->info.bootstrap.b64BitDmaAddress = FALSE;
  dfParameterGetBoolean(pDevCtx->pDevObj->pDrvObj, "PciAddress64Bit", &pDevCtx->info.bootstrap.b64BitDmaAddress);
  pDevCtx->info.bootstrap.dmaDescriptorSize = 32;
  pDevCtx->info.bootstrap.dmaTransferMaxSize = 0x1000000U;
  pDevCtx->info.bootstrap.bWorkAround4kCrossing = (pDevCtx->pDevObj->hardwareId.variant.pci.revision <= 0x01U) ? TRUE : FALSE;
}
