/*
** File: event.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for event registration and waiting.
**
** (C) Copyright Alpha Data 2009
*/

#include <df.h>
#include "device.h"
#include "event.h"
#include <admxrc3/types.h>

typedef enum _State {
  StateInitial = 0,
  StateWaiting = 1,
  StateCompleted = 2
} State;

#if DF_HAS_USER_EVENT_HANDLE
static void
notifyAllEvents(
  Admxrc3DeviceContext* pDevCtx,
  UserEventList* pList,
  unsigned int targetIndex)
{
  UserEventInfo* pInfo;
  DfSpinLockFlags f;

  /* Signal all user events in the list */
  f = dfSpinLockGet(&pList->lock);
  pInfo = pList->pHead;
  while (NULL != pInfo) {
    dfUserEventSignal(pInfo->userEvent);
    pInfo = pInfo->pNext;
  }
  dfSpinLockPut(&pList->lock, f);
}
#endif

static void
completeNotificationWaits(
  Admxrc3DeviceContext* pDevCtx,
  NotificationWaitList* pList,
  unsigned int targetIndex,
  DfIoStatus reason)
{
  DfListNode* pHead;
  DfList complete;
  NotificationTicket* pTicket;
  DfSpinLockFlags f;

  dfDebugPrint(6, ("completeNotificationWaits: pList=%p targetIndex=%u reason=%lu(0x%lx)\n",
    targetIndex, (unsigned long)reason, (unsigned long)reason));

  f = dfSpinLockGet(&pList->lock);
  /* Confiscate the list of tickets */
  dfListConfiscate(&complete, &pList->header);
  dfSpinLockPut(&pList->lock, f);

  while (1) {
    f = dfSpinLockGet(&pList->lock);
    pHead = dfListGetHead(&complete);
    if (pHead == NULL) {
      dfSpinLockPut(&pList->lock, f);
      break;
    }
    dfListRemove(pHead);
    pTicket = DF_CONTAINER_OF(pHead, NotificationTicket, node);
    dfDebugPrint(7, ("completeNotificationWaits: pHead=%p pTicket==%p\n", pHead, pTicket));
    pTicket->state = StateCompleted;
    if (dfRequestClearCallback(pTicket->pRequest)) {
      /* Successfully removed cancel callback; we can complete the request */
      dfSpinLockPut(&pList->lock, f);
      dfRequestComplete(pTicket->pRequest, reason);
      dfPoolFree(&pDevCtx->notification.ticketPool, pTicket);
    } else {
      /* Cancel callback must be running; allow it to complete the request */
      dfSpinLockPut(&pList->lock, f);
    }
  }
}

#if DF_HAS_USER_EVENT_HANDLE
/* Removes the user event list and frees all user event nodes that are associated with 'pClCtx' */
static void
cleanupUserEventList(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  UserEventList* pList)
{
  UserEventInfo* pInfo;
  UserEventInfo* pNext;
  DfSpinLockFlags f;

  f = dfSpinLockGet(&pList->lock);
  pInfo = pList->pHead;
  while (NULL != pInfo) {
    pNext = pInfo->pNext;
    if (pInfo->pClCtx == pClCtx) {
      if (NULL != pInfo->pNext) {
        pInfo->pNext->pPrev = pInfo->pPrev;
      } else {
        pList->pTail = pInfo->pNext;
      }
      if (NULL != pInfo->pPrev) {
        pInfo->pPrev->pNext = pInfo->pNext;
      } else {
        pList->pHead = pInfo->pNext;
      }
      dfUserEventPut(pInfo->userEvent);
      dfFree(pInfo);
    }
    pInfo = pNext;
  }
  dfSpinLockPut(&pList->lock, f);
}
#endif

#if DF_HAS_USER_EVENT_HANDLE
/* Removes from the device and frees all user event nodes that are associated with 'pClCtx' */
void
cleanupUserEvents(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx)
{
  UserEventList* pList;
  unsigned int i;

  for (i = 0; i < MAX_NUM_TARGET_FPGA; i++) {
    pList = &pDevCtx->fpga[i].interrupt.eventList;
    cleanupUserEventList(pDevCtx, pClCtx, pList);
    pList = &pDevCtx->fpga[i].alert.eventList;
    cleanupUserEventList(pDevCtx, pClCtx, pList);
  }
}
#endif

/*
** -----------------------------------------------------------------------------
** Callbacks
** -----------------------------------------------------------------------------
*/

void
fpgaInterruptNotifyCallback(
  CoreNotification* pCoreNotification,
  unsigned int notifyType,
  void* pCallbackContext)
{
  Admxrc3DeviceContext* pDevCtx = (Admxrc3DeviceContext*)pCallbackContext;
  FpgaInterruptNotification* pNotification = (FpgaInterruptNotification*)pCoreNotification;
  unsigned int targetIndex;
  NotificationWaitList* pWaitList;
#if DF_HAS_USER_EVENT_HANDLE
  UserEventList* pEventList;
#endif

  dfDebugPrint(6, ("fpgaInterruptNotifyCallback: pNotification=%p notifyType=%lu pCallbackContext=%p\n",
    pNotification, (unsigned long)notifyType, pCallbackContext));

  targetIndex = pNotification->targetIndex;
  pWaitList = &pDevCtx->fpga[targetIndex].interrupt.waitList;
#if DF_HAS_USER_EVENT_HANDLE
  pEventList = &pDevCtx->fpga[targetIndex].interrupt.eventList;
  notifyAllEvents(pDevCtx, pEventList, pNotification->targetIndex);
#endif
  completeNotificationWaits(pDevCtx, pWaitList, targetIndex, DfIoStatusSuccess);
}

void
fpgaAlertNotifyCallback(
  CoreNotification* pCoreNotification,
  unsigned int notifyType,
  void* pCallbackContext)
{
  Admxrc3DeviceContext* pDevCtx = (Admxrc3DeviceContext*)pCallbackContext;
  FpgaAlertNotification* pNotification = (FpgaAlertNotification*)pCoreNotification;
  unsigned int targetIndex;
  NotificationWaitList* pWaitList;
#if DF_HAS_USER_EVENT_HANDLE
  UserEventList* pEventList;
#endif

  dfDebugPrint(6, ("fpgaAlertNotifyCallback: pNotification=%p notifyType=%lu pCallbackContext=%p\n",
    pNotification, (unsigned long)notifyType, pCallbackContext));

  targetIndex = pNotification->targetIndex;
  pWaitList = &pDevCtx->fpga[targetIndex].interrupt.waitList;
#if DF_HAS_USER_EVENT_HANDLE
  pEventList = &pDevCtx->fpga[targetIndex].alert.eventList;
  notifyAllEvents(pDevCtx, pEventList, pNotification->targetIndex);
#endif
  completeNotificationWaits(pDevCtx, pWaitList, targetIndex, DfIoStatusSuccess);
}

static void
notificationWaitCancelCallback(
  DfClientRequest* pRequest,
  DfIoStatus reason)
{
  NotificationTicket* pTicket = (NotificationTicket*)dfRequestGetContext(pRequest);
  Admxrc3DeviceContext* pDevCtx = pTicket->pDevCtx;
  NotificationWaitList* pList;
  unsigned int notification, targetIndex;
  DfSpinLockFlags f;

  notification = pTicket->notification;
  if (notification >= ADMXRC3_EVENT_FPGAINTERRUPT(0) && notification <= ADMXRC3_EVENT_FPGAINTERRUPT(0xffU)) {
    targetIndex = notification - ADMXRC3_EVENT_FPGAINTERRUPT(0);
    dfAssert(targetIndex < pDevCtx->info.numTargetFpga);
    pList = &pDevCtx->fpga[targetIndex].interrupt.waitList;
  } else if (notification >= ADMXRC3_EVENT_FPGAALERT(0) && notification <= ADMXRC3_EVENT_FPGAALERT(0xffU)) {
    targetIndex = notification - ADMXRC3_EVENT_FPGAALERT(0);
    dfAssert(targetIndex < pDevCtx->info.numTargetFpga);
    pList = &pDevCtx->fpga[targetIndex].alert.waitList;
  } else {
    dfAssert(FALSE);
    return;
  }

  f = dfSpinLockGet(&pList->lock);
  switch (pTicket->state) {
  case StateWaiting:
    dfListRemove(&pTicket->node);
    break;

  case StateCompleted:
    break;

  case StateInitial:
  default:
    dfSpinLockPut(&pList->lock, f);
    dfAssert(FALSE);
    return;
  }
  dfSpinLockPut(&pList->lock, f);

  dfRequestComplete(pTicket->pRequest, reason);
  dfPoolFree(&pDevCtx->notification.ticketPool, pTicket);
}

/*
** -----------------------------------------------------------------------------
** IOCTL handlers
** -----------------------------------------------------------------------------
*/

DfIoStatus
ioctlNotificationWait(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_NOTIFICATIONWAIT* pIoctl = (IOCTLS_ADMXRC3_NOTIFICATIONWAIT*)pBuffer;
  unsigned int notification, targetIndex;
  NotificationWaitList* pList;
  NotificationTicket* pTicket;
  DfSpinLockFlags f;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }

  notification = pIoctl->in.notification;

  /* Get the event list corresponding to 'notification' from the client context */
  if (notification >= ADMXRC3_EVENT_FPGAINTERRUPT(0) && notification <= ADMXRC3_EVENT_FPGAINTERRUPT(0xffU)) {
    targetIndex = notification - ADMXRC3_EVENT_FPGAINTERRUPT(0);
    if (targetIndex >= pDevCtx->info.numTargetFpga) {
      return DfIoStatusError(ADMXRC3_INVALID_INDEX);
    }
    pList = &pDevCtx->fpga[targetIndex].interrupt.waitList;
  } else if (notification >= ADMXRC3_EVENT_FPGAALERT(0) && notification <= ADMXRC3_EVENT_FPGAALERT(0xffU)) {
    targetIndex = notification - ADMXRC3_EVENT_FPGAALERT(0);
    if (targetIndex >= pDevCtx->info.numTargetFpga) {
      return DfIoStatusError(ADMXRC3_INVALID_INDEX);
    }
    pList = &pDevCtx->fpga[targetIndex].alert.waitList;
  } else {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  pTicket = dfPoolAlloc(&pDevCtx->notification.ticketPool, NotificationTicket);
  if (NULL == pTicket) {
    return DfIoStatusError(ADMXRC3_NO_MEMORY);
  }
  pTicket->pDevCtx = pDevCtx;
  pTicket->pRequest = pRequest;
  pTicket->state = StateInitial;
  pTicket->notification = notification;
  dfRequestSetContext(pRequest, pTicket);

  f = dfSpinLockGet(&pList->lock);
  if (!dfRequestSetCallback(pRequest, notificationWaitCancelCallback)) {
    dfSpinLockPut(&pList->lock, f);
    dfPoolFree(&pDevCtx->notification.ticketPool, pTicket);
    return DfIoStatusError(ADMXRC3_CANCELLED);
  }
  dfListAddToTail(&pList->header, &pTicket->node);
  pTicket->state = StateWaiting;
  dfSpinLockPut(&pList->lock, f);

  dfRequestPending(pRequest);
  return DfIoStatusDeferred;
}

#if DF_HAS_USER_EVENT_HANDLE
DfIoStatus
ioctlRegisterEvent(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_REGISTEREVENT* pIoctl = (IOCTLS_ADMXRC3_REGISTEREVENT*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  DfSpinLockFlags f;
  unsigned int notification, targetIndex;
  user_event_handle_t hEvent;
  DfUserEvent userEvent;
  UserEventInfo* pInfo;
  UserEventList* pList;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }

  notification = pIoctl->in.notification;
  hEvent = pIoctl->in.hEvent;

  /* Get the event list corresponding to 'notification' from the client context */
  if (notification >= ADMXRC3_EVENT_FPGAINTERRUPT(0) && notification <= ADMXRC3_EVENT_FPGAINTERRUPT(0xffU)) {
    targetIndex = notification - ADMXRC3_EVENT_FPGAINTERRUPT(0);
    if (targetIndex >= pDevCtx->info.numTargetFpga) {
      return DfIoStatusError(ADMXRC3_INVALID_INDEX);
    }
    pList = &pDevCtx->fpga[targetIndex].interrupt.eventList;
  } else if (notification >= ADMXRC3_EVENT_FPGAALERT(0) && notification <= ADMXRC3_EVENT_FPGAALERT(0xffU)) {
    targetIndex = notification - ADMXRC3_EVENT_FPGAALERT(0);
    if (targetIndex >= pDevCtx->info.numTargetFpga) {
      return DfIoStatusError(ADMXRC3_INVALID_INDEX);
    }
    pList = &pDevCtx->fpga[targetIndex].alert.eventList;
  } else {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  /* Convert the user-mode event handle to a kernel user-mode event */
  if (!dfUserEventGet(hEvent, &userEvent)) {
    return DfIoStatusError(ADMXRC3_INVALID_HANDLE);
  }

  /* Create a user-mode event information node */
  pInfo = (UserEventInfo*)dfMalloc(sizeof(UserEventInfo));
  if (NULL == pInfo) {
    dfUserEventPut(userEvent);
    return DfIoStatusError(ADMXRC3_NO_MEMORY);
  }
  pInfo->pClCtx = pClCtx;
  pInfo->hUserEvent = hEvent;
  pInfo->userEvent = userEvent;

  /* Add node to user-mode event list for client context */
  f = dfSpinLockGet(&pList->lock);
  pInfo->pNext = NULL;
  pInfo->pPrev = pList->pHead;
  if (NULL == pList->pHead) {
    pList->pHead = pInfo;
    pList->pTail = pInfo;
  } else {
    pList->pTail->pNext = pInfo;
    pList->pTail = pInfo;
  }
  dfSpinLockPut(&pList->lock, f);

  return status;
}
#endif

#if DF_HAS_USER_EVENT_HANDLE && DF_NEED_THUNK
DfIoStatus
ioctlRegisterEventThunk(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_REGISTEREVENT* pIoctl32 = (IOCTLS32_ADMXRC3_REGISTEREVENT*)pBuffer;
  IOCTLS_ADMXRC3_REGISTEREVENT ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.notification = pIoctl32->in.notification;
  ioctl.in.hEvent = dfThunkHandle(pIoctl32->in.hEvent);
  return ioctlRegisterEvent(pDevCtx, pClCtx, &ioctl, sizeof(ioctl.in));
}
#endif

#if DF_HAS_USER_EVENT_HANDLE
DfIoStatus
ioctlUnregisterEvent(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_UNREGISTEREVENT* pIoctl = (IOCTLS_ADMXRC3_UNREGISTEREVENT*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  DfSpinLockFlags f;
  unsigned int notification, targetIndex;
  user_event_handle_t hEvent;
  UserEventList* pList;
  UserEventInfo* pInfo;
  UserEventInfo* pNext;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }

  notification = pIoctl->in.notification;
  hEvent = pIoctl->in.hEvent;

  /* Get the user-mode event list corresponding to 'notification' from the client context */
  if (notification >= ADMXRC3_EVENT_FPGAINTERRUPT(0) && notification <= ADMXRC3_EVENT_FPGAINTERRUPT(0xffU)) {
    targetIndex = notification - ADMXRC3_EVENT_FPGAINTERRUPT(0);
    if (targetIndex >= pDevCtx->info.numTargetFpga) {
      return DfIoStatusError(ADMXRC3_INVALID_INDEX);
    }
    pList = &pDevCtx->fpga[targetIndex].interrupt.eventList;
  } else if (notification >= ADMXRC3_EVENT_FPGAALERT(0) && notification <= ADMXRC3_EVENT_FPGAALERT(0xffU)) {
    targetIndex = notification - ADMXRC3_EVENT_FPGAALERT(0);
    if (targetIndex >= pDevCtx->info.numTargetFpga) {
      return DfIoStatusError(ADMXRC3_INVALID_INDEX);
    }
    pList = &pDevCtx->fpga[targetIndex].alert.eventList;
  } else {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  /* Search the user-mode event list for a node matching 'hEvent', and remove it from the list if found */
  f = dfSpinLockGet(&pList->lock);
  pInfo = pList->pHead;
  while (NULL != pInfo) {
    pNext = pInfo->pNext;
    if (pInfo->pClCtx == pClCtx && pInfo->hUserEvent == hEvent) {
      /* Found the node; remove it from the event list */
      if (NULL != pInfo->pNext) {
        pInfo->pNext->pPrev = pInfo->pPrev;
      } else {
        pList->pTail = pInfo->pPrev;
      }
      if (NULL != pInfo->pPrev) {
        pInfo->pPrev->pNext = pInfo->pNext;
      } else {
        pList->pHead = pInfo->pNext;
      }
      break;
    }
    pInfo = pNext;
  }
  dfSpinLockPut(&pList->lock, f);

  /* Nothing was found in the list; hence 'hEvent' must be considered invalid */
  if (NULL == pInfo) {
    return DfIoStatusError(ADMXRC3_INVALID_HANDLE);
  }

  /* Found a match; release the kernel user-mode event & free the list node */
  dfUserEventPut(pInfo->userEvent);
  dfFree(pInfo);

  return status;
}
#endif

#if DF_HAS_USER_EVENT_HANDLE && DF_NEED_THUNK
DfIoStatus
ioctlUnregisterEventThunk(
	Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_UNREGISTEREVENT* pIoctl32 = (IOCTLS32_ADMXRC3_UNREGISTEREVENT*)pBuffer;
  IOCTLS_ADMXRC3_UNREGISTEREVENT ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.notification = pIoctl32->in.notification;
  ioctl.in.hEvent = dfThunkHandle(pIoctl32->in.hEvent);
  return ioctlUnregisterEvent(pDevCtx, pClCtx, &ioctl, sizeof(ioctl.in));
}
#endif
