/*
** File: vpd.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for reading and writing VPD (vital
**          Product Data).
**
** (C) Copyright Alpha Data 2010
*/

#include <df.h>
#include <core/coreif.h>
#include "device.h"
#include "vpd.h"

#define MAX_CHUNK (128)

static DfIoStatus
mapAcquireStatus(
  CoreAcquireStatus acqStatus)
{
  switch (acqStatus) {
  case CoreAcquireSuccess:
    return DfIoStatusSuccess;

  case CoreAcquireMustWait:
    return DfIoStatusError(ADMXRC3_MUST_WAIT);

  case CoreAcquireCancelled:
    return DfIoStatusError(ADMXRC3_CANCELLED);

  case CoreAcquireRemoved:
  case CoreAcquireFailed:
    return DfIoStatusError(ADMXRC3_HARDWARE_ERROR);

  case CoreAcquireInvalidIndex:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);

  case CoreAcquireInvalid:
  case CoreAcquireUnexpectedError:
  case CoreAcquireWaiting:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

static DfIoStatus
mapVpdStatus(
  CoreVpdStatus vpdStatus)
{
  switch (vpdStatus) {
  case CoreVpdSuccess:
    return DfIoStatusSuccess;
  case CoreVpdNotPresent:
  case CoreVpdTimeout:
  case CoreVpdVerifyFailed:
  case CoreVpdWriteFailed:
    return DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
  case CoreVpdInvalidRegion:
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  case CoreVpdInterrupted:
    return DfIoStatusError(ADMXRC3_CANCELLED);
  case CoreVpdInvalidParameter:
  case CoreVpdGeneralFailure:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

DfIoStatus
ioctlReadVpd(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
#if defined(ADMXRC3_CHECK_FLAGS)
  static const uint32_t legalFlagsMask = 0;
#endif
  IOCTLS_ADMXRC3_READVPD* pIoctl = (IOCTLS_ADMXRC3_READVPD*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  CoreAcquireStatus acqStatus;
  CoreVpdStatus vpdStatus;
  CoreTicket ticket;
  uint8_t stagingBuffer[MAX_CHUNK];
  uint32_t flags;
  size_t offset;
  size_t length, chunk;
  uint8_t* pUserBuf;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  flags = pIoctl->in.flags;
  offset = pIoctl->in.offset;
  length = pIoctl->in.length;
  pUserBuf = pIoctl->in.pBuffer;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif

  ticket.count = 1;
  ticket.resources[0].code = CORE_RESOURCE_VPD;
  ticket.resources[0].index = 0;
  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &ticket);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }
  acqStatus = pDevCtx->coreInterface.pAcquireSync(pDevCtx->pCoreContext, &ticket, 0);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }

  DF_USER_SPACE_ACCESS_BEGIN

  dfUserBufferValidate(pUserBuf, length, TRUE);

  while (length) {
    chunk = (length > MAX_CHUNK) ? MAX_CHUNK : length;
    vpdStatus = pDevCtx->coreInterface.pVpdRead(pDevCtx->pCoreContext, &ticket, FALSE, offset, chunk, stagingBuffer);
    if (CoreVpdSuccess != vpdStatus) {
      status = mapVpdStatus(vpdStatus);
      break;
    }
    dfCopyKU(pUserBuf, stagingBuffer, chunk);
    pUserBuf += chunk;
    offset += chunk;
    length -= chunk;
  }

  DF_USER_SPACE_ACCESS_END({
    status = DfIoStatusError(ADMXRC3_INVALID_BUFFER);
  })

  pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &ticket);

  return status;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlReadVpdThunk(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_READVPD* pIoctl32 = (IOCTLS32_ADMXRC3_READVPD*)pBuffer;
  IOCTLS_ADMXRC3_READVPD ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.offset = (size_t)pIoctl32->in.offset;
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.pBuffer = dfThunkPtr(pIoctl32->in.pBuffer);
  return ioctlReadVpd(pDevCtx, pClCtx, pReq, &ioctl, sizeof(ioctl.in));
}
#endif

DfIoStatus
ioctlWriteVpd(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
#if defined(ADMXRC3_CHECK_FLAGS)
  static const uint32_t legalFlagsMask = 0;
#endif
  IOCTLS_ADMXRC3_WRITEVPD* pIoctl = (IOCTLS_ADMXRC3_WRITEVPD*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  CoreAcquireStatus acqStatus;
  CoreVpdStatus vpdStatus;
  CoreTicket ticket;
  uint8_t stagingBuffer[MAX_CHUNK];
  uint32_t flags;
  size_t offset;
  size_t length, chunk;
  const uint8_t* pUserBuf;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  flags = pIoctl->in.flags;
  offset = pIoctl->in.offset;
  length = pIoctl->in.length;
  pUserBuf = (const uint8_t*)pIoctl->in.pBuffer;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif

  ticket.count = 1;
  ticket.resources[0].code = CORE_RESOURCE_VPD;
  ticket.resources[0].index = 0;
  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &ticket);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }
  acqStatus = pDevCtx->coreInterface.pAcquireSync(pDevCtx->pCoreContext, &ticket, 0);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }

  DF_USER_SPACE_ACCESS_BEGIN

  dfUserBufferValidate(pUserBuf, length, FALSE);

  while (length) {
    chunk = (length > MAX_CHUNK) ? MAX_CHUNK : length;
    dfCopyUK(stagingBuffer, pUserBuf, chunk);
    vpdStatus = pDevCtx->coreInterface.pVpdWrite(pDevCtx->pCoreContext, &ticket, FALSE, offset, chunk, stagingBuffer);
    if (CoreVpdSuccess != vpdStatus) {
      status = mapVpdStatus(vpdStatus);
      break;
    }
    pUserBuf += chunk;
    length -= chunk;
    offset += chunk;
  }

  DF_USER_SPACE_ACCESS_END({
    status = DfIoStatusError(ADMXRC3_INVALID_BUFFER);
  })

  pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &ticket);

  return status;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlWriteVpdThunk(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_WRITEVPD* pIoctl32 = (IOCTLS32_ADMXRC3_WRITEVPD*)pBuffer;
  IOCTLS_ADMXRC3_WRITEVPD ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.offset = (size_t)pIoctl32->in.offset;
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.pBuffer = dfThunkConstPtr(pIoctl32->in.pBuffer);
  return ioctlWriteVpd(pDevCtx, pClCtx, pReq, &ioctl, sizeof(ioctl.in));
}
#endif

